@@ -46,6 +46,7 @@ debug_level = 0
 top_window = None
 last_rsc_view_page_num = 0	
 last_compound_view_page_num = 0
+last_cluster_view_page_num = 0
 
 def log(string) :
 	syslog.syslog(string)
@@ -134,13 +135,25 @@ def msgbox(msg) :
 	dialog.destroy()
 
 
-def confirmbox(msg, button=(gtk.STOCK_YES, gtk.RESPONSE_YES, gtk.STOCK_NO, gtk.RESPONSE_NO)) :
+def confirmbox(msg, button=(gtk.STOCK_YES, gtk.RESPONSE_YES, gtk.STOCK_NO, gtk.RESPONSE_NO), title ="") :
 	global top_window
 	dialog = gtk.Dialog(_("Confirm"),  top_window, gtk.DIALOG_MODAL,button)
-	label = gtk.Label(msg)
-	dialog.vbox.add(label)
+	im=gtk.Image()
+	im.set_from_stock(gtk.STOCK_DIALOG_QUESTION, gtk.ICON_SIZE_DIALOG)
+	frame = gtk.Frame()
+	confirm_title = gtk.Label()
+	confirm_title.set_markup('<b>'+title+'</b>')
+	frame.set_label_widget(confirm_title)
+	confirm_msg =  gtk.Label()
+	confirm_msg.set_text(msg)
+	dialog.vbox.pack_start(frame)
+	hb=gtk.HBox()
+	hb.pack_start(im)
+	hb.pack_start(confirm_msg)
+	frame.add(hb)
+	dialog.vbox.resize_children()
 	dialog.set_default_response(gtk.RESPONSE_YES)
-	label.show()
+	dialog.show_all()
 	save_top_window = top_window
 	top_window = dialog
 	ret = dialog.run()
@@ -189,11 +202,13 @@ def kvbox(title, fields, auto_fill = None) :
 			table.attach(entry, 1,2, row, row+1)
 
 	if auto_fill != None :
-		if title == _("Add Parameter") :
+		if title == _("Add Parameter") and combos.has_key("name"):
 			combos["name"].connect("changed", auto_fill, widgets)
 		elif title == _("Add Operation") :
-			combos["name"].connect("changed", auto_fill, widgets, "name")
-			combos["role"].connect("changed", auto_fill, widgets, "role")
+			if combos.has_key("name"):
+				combos["name"].connect("changed", auto_fill, widgets, "name")
+			if combos.has_key("role"):
+				combos["role"].connect("changed", auto_fill, widgets, "role")
 	
 	dialog.vbox.pack_start(table)
 	dialog.vbox.show_all()
@@ -546,7 +561,7 @@ class Tree :
 				iter_status = 1
 		elif type in [_("group"),_("resources"),_("clone"),_("master")]:
 			iter_status = 0
-		elif type in [_("Orders"),_("Places"), _("Colocations")] :
+		elif type in [_("Orders"),_("Locations"), _("Colocations")] :
 			iter_status = -1
 		else :
 			iter_status = -2
@@ -619,7 +634,7 @@ class Tree :
 			return
 		
 		config = manager.get_cluster_config()
-		if config["have_quorum"] == "True" :
+		if config.has_key("have_quorum") and config["have_quorum"] == "true" :
 			status = _("with quorum")
 		else :
 			status = _("without quorum")
@@ -636,10 +651,10 @@ class Tree :
 
 		constraints_root = self.store.append(root, [_("Constraints"),"", "constraints"])
 		
-		places_root = self.store.append(constraints_root, [_("Places"),"", "places"])
-		places = manager.get_constraints("rsc_location")
-		for place in places :
-			self.store.append(places_root,[place["id"], "", _("place")])
+		locations_root = self.store.append(constraints_root, [_("Locations"),"", "locations"])
+		locations = manager.get_constraints("rsc_location")
+		for location in locations :
+			self.store.append(locations_root,[location["id"], "", _("location")])
 		
 		orders_root = self.store.append(constraints_root, [_("Orders"),"", "orders"])
 		orders = manager.get_constraints("rsc_order")
@@ -722,6 +737,8 @@ class View :
 			self.glade.get_widget("apply").connect("clicked", self.on_apply)
 		if self.glade.get_widget("reset") != None :
 			self.glade.get_widget("reset").connect("clicked", self.on_reset)
+		if self.glade.get_widget("default") != None :
+			self.glade.get_widget("default").connect("clicked", self.on_default)
 			
 	def update(self) :
 		if self.glade.get_widget("apply") != None :
@@ -742,6 +759,9 @@ class View :
 		self.glade.get_widget("apply").set_property("sensitive", False)
 		self.glade.get_widget("reset").set_property("sensitive", False)
 		self.changed = False
+
+	def on_default(self, widget):
+		self.on_changed(widget)
 		
 	def on_after_show(self) :
 		pass	
@@ -778,25 +798,60 @@ class NodeView(View) :
 		
 class ClusterView(View) :
 	name = "clusterview"
+	pengine_metadata = None
+	crmd_metadata = None
+	pengine_widgets = None
+	crmd_widgets = None
 	
 	def on_apply(self, widget):
 		glade = self.glade
-		new_crm_config = {}
-		new_crm_config["default-action-timeout"] = glade.get_widget("transition_timeout").get_text()
-		new_crm_config["symmetric-cluster"] = str(glade.get_widget("symmetric_cluster").get_active())
-		new_crm_config["stonith-enabled"] = str(glade.get_widget("stonith_enabled").get_active())
-		new_crm_config["no-quorum-policy"] = glade.get_widget("no_quorum_policy").child.get_text()
-		new_crm_config["default-resource-stickiness"] = \
-				glade.get_widget("default_resource_stickiness").child.get_text()
-		new_crm_config["default-resource-failure-stickiness"] = \
-				glade.get_widget("default_resource_failure_stickiness").child.get_text()
-		manager.update_crm_config(new_crm_config)
 		
+		current_page = glade.get_widget("notebook").get_current_page()
+		if current_page == 1:
+			(widgets, metadata) = (self.pengine_widgets, self.pengine_metadata)
+		elif current_page == 2:
+			(widgets, metadata) = (self.crmd_widgets, self.crmd_metadata)
+
+		new_crm_config = {}
+		for key in widgets["combos"].keys():
+			new_crm_config[key] = widgets["combos"][key].child.get_text()
+		for key in widgets["checkbuttons"].keys():
+			if widgets["checkbuttons"][key].get_active():
+				new_crm_config[key] = "true"
+			else:
+				new_crm_config[key] = "false"
+		for key in widgets["entries"].keys():
+			new_crm_config[key] = widgets["entries"][key].get_text()
+		manager.update_crm_config(metadata, new_crm_config)
+
 		View.on_apply(self, widget)
+
+	def on_default(self, widget):
+		current_page = self.glade.get_widget("notebook").get_current_page()
+		if current_page == 1:
+			(widgets, metadata) = (self.pengine_widgets, self.pengine_metadata)
+		elif current_page == 2:
+			(widgets, metadata) = (self.crmd_widgets, self.crmd_metadata)
+
+		default_crm_config = {}
+		for parameter in metadata.parameters:
+			default_crm_config[parameter["name"]] = str(parameter["content"]["default"])
+
+		if default_crm_config != {}:
+			for key in widgets["combos"].keys():
+				widgets["combos"][key].child.set_text(default_crm_config[key])
+				widgets["combos"][key].child.set_editable(False)
+			for key in widgets["checkbuttons"].keys():
+				widgets["checkbuttons"][key].set_active("true" == default_crm_config[key])
+			for key in widgets["entries"].keys():
+				widgets["entries"][key].set_text(default_crm_config[key])
+
+		View.on_default(self, widget)
 	
 	def update(self) :
 		glade = self.glade
 		config = manager.get_cluster_config()
+
 		if config != None :
 			glade.get_widget("hbversion").set_text(config["hbversion"])
 			glade.get_widget("debug").set_text(str(config["debug"]))
@@ -806,14 +861,18 @@ class ClusterView(View) :
 			glade.get_widget("deadtime").set_text(config["deadtime"])
 			glade.get_widget("initdead").set_text(config["initdead"])
 
-			glade.get_widget("transition_timeout").set_text(config["default-action-timeout"])
-			glade.get_widget("symmetric_cluster").set_active("True" == config["symmetric-cluster"])
-			glade.get_widget("stonith_enabled").set_active("True" == config["stonith-enabled"])
-			glade.get_widget("no_quorum_policy").child.set_text(config["no-quorum-policy"])
-			glade.get_widget("no_quorum_policy").child.set_editable(False)
-			glade.get_widget("default_resource_stickiness").child.set_text(config["default-resource-stickiness"])
-			glade.get_widget("default_resource_failure_stickiness").child.set_text \
-							 (config["default-resource-failure-stickiness"])
+		for (widgets, metadata) in ((self.pengine_widgets, self.pengine_metadata),
+						(self.crmd_widgets, self.crmd_metadata)):
+			if metadata != None and widgets != None:
+				crm_config = manager.get_crm_config(metadata)
+				if crm_config != {}:
+					for key in widgets["combos"].keys():
+						widgets["combos"][key].child.set_text(crm_config[key])
+						widgets["combos"][key].child.set_editable(False)
+					for key in widgets["checkbuttons"].keys():
+						widgets["checkbuttons"][key].set_active("true" == crm_config[key])
+					for key in widgets["entries"].keys():
+						widgets["entries"][key].set_text(crm_config[key])
 
 		View.update(self)
 	
@@ -825,26 +884,109 @@ class ClusterView(View) :
 		make_label_active(self.glade.get_widget("l_warntime"),_("Warning Alive:"),"www.linux-ha.org")
 		make_label_active(self.glade.get_widget("l_deadtime"),_("Dead Time:"),"www.linux-ha.org")
 		make_label_active(self.glade.get_widget("l_initdead"),_("Initial Dead Time:"),"www.linux-ha.org")
-		make_label_active(self.glade.get_widget("l_transtimeout"),_("Transition Timeout:"),"www.linux-ha.org")
-		make_label_active(self.glade.get_widget("l_rscstickiness"),_("Resource Stickiness:"),"www.linux-ha.org")
-		make_label_active(self.glade.get_widget("l_noquorum"),_("No Quorum Policy:"),"www.linux-ha.org")
-		make_label_active(self.glade.get_widget("l_symmetric"),_("Symmetric Cluster"),"www.linux-ha.org")
-		make_label_active(self.glade.get_widget("l_stonith_enabled"),_("Stonith Enabled"),"www.linux-ha.org")
-		make_label_active(self.glade.get_widget \
-			("l_rscfailstickiness"),_("Resource Failure Stickiness:"),"www.linux-ha.org")
+
+		name_cap = lambda name: name.replace("-"," ").replace("_"," ").title().replace("Crmd ","CRMD ").\
+						replace("Pe ", "PE ").replace("Dc ","DC ")
+
+		for widgets in (self.pengine_widgets, self.crmd_widgets):
+			if widgets != None:
+				for key in widgets["labels"].keys():
+					make_label_active(widgets["labels"][key], 
+								_(name_cap(key)+":"),
+								"www.linux-ha.org")
 							
+	def on_notebook_event(self, widget, page, page_num) :
+		global last_cluster_view_page_num
+
+		if self.changed == True:
+			ret = confirmbox(_("There are unsaved changes in current page.")+"\n"+_("Apply the changes?"),
+					(gtk.STOCK_YES, gtk.RESPONSE_YES, 
+					gtk.STOCK_NO, gtk.RESPONSE_NO))
+			if ret == gtk.RESPONSE_YES :
+				self.on_apply(widget)
+			if ret == gtk.RESPONSE_NO :
+				self.update()
+
+		if page_num == 0:
+			self.glade.get_widget("default").set_property("sensitive", False)
+		else:
+			self.glade.get_widget("default").set_property("sensitive", True)
+
+		last_cluster_view_page_num = page_num
+
 	def __init__(self) :
 		View.__init__(self)
-		
 		glade = self.glade
-		glade.get_widget("transition_timeout").connect("changed", self.on_changed)
-		glade.get_widget("symmetric_cluster").connect("toggled", self.on_changed)
-		glade.get_widget("stonith_enabled").connect("toggled", self.on_changed)
-		glade.get_widget("no_quorum_policy").connect("changed", self.on_changed)
-		glade.get_widget("default_resource_stickiness").connect("changed", self.on_changed)
-		glade.get_widget("default_resource_failure_stickiness").connect("changed", self.on_changed)
+
+		self.pengine_metadata = manager.get_crm_metadata("pengine")
+		self.crmd_metadata = manager.get_crm_metadata("crmd")
+
+		if self.pengine_metadata != None:
+			self.pengine_widgets = self.render_cluster_conf(self.pengine_metadata)
+		if self.crmd_metadata != None:
+			self.crmd_widgets = self.render_cluster_conf(self.crmd_metadata)
+
+		notebook = glade.get_widget("notebook")
+		notebook.connect("switch-page", self.on_notebook_event)
+		notebook.set_current_page(last_cluster_view_page_num)
+
+		if last_cluster_view_page_num == 0 :
+			glade.get_widget("default").set_property("sensitive", False)
+
 		self.update()
-		
+
+
+
+	def render_cluster_conf(self, metadata):
+		if metadata.name == "Policy Engine":
+			vbox = self.glade.get_widget("vbox_pengine")
+		elif metadata.name == "CRM Daemon":
+			vbox = self.glade.get_widget("vbox_crmd")
+
+		widgets = {}
+		labels = {}
+		combos = {}
+		checkbuttons = {}
+		entries = {}
+		table = gtk.Table(2, 1, False)
+		parameters = metadata.parameters
+
+		for row in range(len(parameters)):
+			label = gtk.Label(parameters[row]["name"]+":")
+			label.set_alignment(1, 0.5)
+			eventbox = gtk.EventBox()
+			eventbox.add(label)
+			table.attach(eventbox, 0, 1, row, row+1, xpadding = 5)
+			labels[parameters[row]["name"]] = label
+
+			if parameters[row]["content"].has_key("values"):
+				combo = gtk.combo_box_entry_new_text()
+				for option in parameters[row]["content"]["values"]:
+					combo.append_text(option)
+				combo.connect("changed", self.on_changed)
+				table.attach(combo, 1, 2, row, row+1, xpadding = 5)
+				combos[parameters[row]["name"]] = combo
+			elif parameters[row]["content"]["type"] == "boolean":
+				checkbutton = gtk.CheckButton()
+				checkbutton.connect("toggled", self.on_changed)
+				table.attach(checkbutton, 1, 2, row, row+1, xpadding = 2)
+				checkbuttons[parameters[row]["name"]] = checkbutton
+			else:
+				entry = gtk.Entry()
+				entry.connect("changed", self.on_changed)
+				table.attach(entry, 1,2, row, row+1, xpadding = 5)
+				entries[parameters[row]["name"]] = entry 
+	
+		vbox.pack_start(table)
+		vbox.show_all()
+
+		widgets["labels"] = labels
+		widgets["combos"] = combos
+		widgets["checkbuttons"] = checkbuttons
+		widgets["entries"] = entries
+	
+		return widgets
+
 class OrderView(View) :
 	name = "orderview"
 	save_order = None
@@ -927,32 +1069,47 @@ class ColocationView(View) :
 		glade.get_widget("score").connect("changed", self.on_changed)
 		self.update()		
 				
-class PlaceView(View) :
-	name = "placeview"
+class LocationView(View) :
+	name = "locationview"
 	expr_list = None
-	place_save = None
+	save_location = None
 	
 	def on_apply(self, widget):
-		place = self.place_save
-		place["exprs"] = self.expr_list.get_data()
-		manager.update_constraint("rsc_location",place)
+		glade = self.glade
+		location = self.save_location
+		location["rsc"] = glade.get_widget("resource").child.get_text()
+		location["score"] = glade.get_widget("score").child.get_text()
+		location["boolean_op"] = glade.get_widget("boolean_op").child.get_text()
+		location["exprs"] = self.expr_list.get_data()
+		manager.update_constraint("rsc_location",location)
 		View.on_apply(self, widget)
 							   
 	def update(self) :
+		glade = self.glade
 		self.expr_list.clear()
-		place = manager.get_constraint("rsc_location", self.param)
-		self.place_save = place
-		if place != None :
-			self.glade.get_widget("id").set_text(place["id"])
-			self.glade.get_widget("resource").set_text(place["rsc"])
-			self.glade.get_widget("score").set_text(place["score"])
-			for expr in place["exprs"] :
+		location = manager.get_constraint("rsc_location", self.param)
+		if location != None :
+			glade.get_widget("id").set_text(location["id"])
+			glade.get_widget("resource").child.set_text(location["rsc"])
+			glade.get_widget("resource").child.set_editable(False)
+			glade.get_widget("score").child.set_text(location["score"])
+			glade.get_widget("boolean_op").child.set_text(location["boolean_op"])
+			glade.get_widget("boolean_op").child.set_editable(False)
+			for expr in location["exprs"] :
 				self.expr_list.insert(expr)
+			store = gtk.ListStore(str)
+			for rsc in manager.get_all_rsc_id():
+				store.append([rsc])
+			glade.get_widget("resource").set_model(store)
+			self.save_location = location
 		View.update(self)
 		
 	def __init__(self, rsc) :
 		View.__init__(self, rsc)
 		glade = self.glade
+		glade.get_widget("resource").connect("changed", self.on_changed)
+		glade.get_widget("score").connect("changed", self.on_changed)
+		glade.get_widget("boolean_op").connect("changed", self.on_changed)
 		self.expr_list = ListWithAddDel(glade.get_widget("expr_list"),
 			glade.get_widget("add_expr"),
 			glade.get_widget("del_expr"),
@@ -968,12 +1125,13 @@ class PlaceView(View) :
 class RscView(View) :
 
 	name = "rscview"
+	metaattr_list = None
 	param_list = None
 	op_list = None
+	metaattrs_save = None
 	params_save = None
 	ops_save = None
 	
-	attr_list = None
 	attrs_save = None
 	meta = None
 	def on_apply(self, widget):
@@ -989,19 +1147,20 @@ class RscView(View) :
 					["id", "name","description","interval","timeout",
 				 	"start_delay","disabled","role","prereq","on_fail"]);
 		
-		nv = self.attr_list.get_data()
-		for n in self.attrs_save.keys():
-			if n not in ["id", "class","type","provider"] :
-				if n not in nv.keys() :
-					manager.update_rsc_attr(self.param, n, "#default")
-		for n in nv.keys() :
-			manager.update_rsc_attr(self.param, n, nv[n])
+		nv = self.metaattr_list.get_data()
+		new_metaattrs = []
+		for n in nv.keys():
+			new_metaattrs.append({"id": self.param + "_metaattr_" + n,
+						"name": n, "value": nv[n]})
+		manager.update_attrs("up_rsc_metaattrs\n"+self.param,"del_rsc_metaattr",
+				self.metaattrs_save, new_metaattrs, ["id","name","value"]);
 				   
 	def update(self) :
 		glade = self.glade
-		(attrs, running_on, params, ops) = \
+		(attrs, running_on, metaattrs, params, ops) = \
 			manager.get_rsc_info(self.param)
 
+		self.metaattrs_save = metaattrs
 		self.params_save = params
 		self.ops_save = ops
 		
@@ -1013,6 +1172,11 @@ class RscView(View) :
 
 		glade.get_widget("runningnodes").set_text(str(running_on))
 						
+		self.metaattr_list.clear()
+		if metaattrs != [] :
+			for metaattr in metaattrs:
+				self.metaattr_list.insert(metaattr["name"], metaattr["value"])
+
 		self.param_list.clear()
 		if params != [] :
 			for param in params:
@@ -1023,11 +1187,7 @@ class RscView(View) :
 			for op in ops:
 				self.op_list.insert(op)
 		
-		self.attr_list.clear()
 		attrs = manager.get_rsc_attrs(self.param)
-		for key in attrs.keys() :
-			if key not in ["id", "class","type","provider"] :
-				self.attr_list.insert(key, attrs[key])
 		self.attrs_save = attrs
 		View.update(self)
 	def on_notebook_event(self, widget, page, page_num) :
@@ -1070,7 +1230,7 @@ class RscView(View) :
 		global last_rsc_view_page_num
 		glade = self.glade
 		
-		(attrs, running_on, params, ops) = manager.get_rsc_info(self.param)
+		(attrs, running_on, metaattrs, params, ops) = manager.get_rsc_info(self.param)
 		meta = manager.get_rsc_meta(attrs["class"], attrs["type"], attrs["provider"])
 		self.meta = meta
 		param_options = None
@@ -1111,9 +1271,10 @@ class RscView(View) :
 			"restart_type":["ignore","restart"],
 			"multiple_active":["stop_start","stop_only","block"],
 			"resource_stickiness":[],
-            "resource_failure_stickiness":[],            
-			"description":[]}
-		self.attr_list = NVList(glade.get_widget("attributes"),
+            		"resource_failure_stickiness":[],            
+			"description":[],
+			"target_role":["started", "stopped", "master", "slave", "default"]}
+		self.metaattr_list = NVList(glade.get_widget("attributes"),
 					glade.get_widget("addattr"), 
 					glade.get_widget("delattr"),
 					options_dict,
@@ -1125,9 +1286,10 @@ class RscView(View) :
 		
 class CompoundRscView(View) :
 	name = "compoundrscview"
+	metaattr_list = None
 	param_list = None
+	metaattrs_save = None
 	params_save = None
-	attr_list = None
 	attrs_save = None
 			
 	def on_apply(self, widget):
@@ -1138,32 +1300,35 @@ class CompoundRscView(View) :
 		manager.update_attrs("up_rsc_params\n"+rid,"del_rsc_param",
 				   self.params_save, new_params, ["id","name","value"]);
 				   
-		nv = self.attr_list.get_data()
-		for n in self.attrs_save.keys():
-			if n not in ["id", "class","type","provider"] :
-				if n not in nv.keys() :
-					manager.update_rsc_attr(rid, n, "#default")
-		for n in nv.keys() :
-			manager.update_rsc_attr(rid, n, nv[n])
+		nv = self.metaattr_list.get_data()
+		new_metaattrs = []
+		for n in nv.keys():
+			new_metaattrs.append({"id": rid + "_metaattr_" + n,
+						"name": n, "value": nv[n]})
+		manager.update_attrs("up_rsc_metaattrs\n"+rid,"del_rsc_metaattr",
+				self.metaattrs_save, new_metaattrs, ["id","name","value"]);
 		
 	def update(self) :
 		glade = self.glade
 		(rid,rtype) = self.param
+		metaattrs = manager.get_rsc_metaattrs(rid)
+		self.metaattrs_save = metaattrs
 		params = manager.get_rsc_params(rid)
 		self.params_save = params
 		glade.get_widget("id").set_text(rid)
 		glade.get_widget("type").set_text(rtype)
-						
+
+		self.metaattr_list.clear()
+		if metaattrs != [] :
+			for metaattr in metaattrs:
+				self.metaattr_list.insert(metaattr["name"], metaattr["value"])
+			
 		self.param_list.clear()
 		if params != [] :
 			for param in params:
 				self.param_list.insert(param)
 		
-		self.attr_list.clear()
 		attrs = manager.get_rsc_attrs(rid)
-		for key in attrs.keys() :
-			if key != "id" :
-				self.attr_list.insert(key, attrs[key])
 		self.attrs_save = attrs
 		View.update(self)
 	
@@ -1188,15 +1353,20 @@ class CompoundRscView(View) :
 			"restart_type":["ignore","restart"],
 			"multiple_active":["stop_start","stop_only","block"],
 			"ordered":["false", "true"], 
-			"collocated":["true","false"], 
-			"notify":["false", "true"],
-			"globally_unique":["true","false"],
-			"interleave":["false","true"],
 			"resource_stickiness":[],
-            "resource_failure_stickiness":[],
-			"description":[]}
+			"resource_failure_stickiness":[],
+			"description":[],
+			"target_role":["started", "stopped", "master", "slave", "default"]}
 		
-		self.attr_list = NVList(glade.get_widget("attributes"),
+		if self.param[1] == _("group"):
+			options_dict["collocated"] = ["true","false"] 
+
+		if self.param[1] in [_("clone"), _("master")]:
+			options_dict["notify"] = ["false", "true"]
+			options_dict["globally_unique"] = ["true","false"]
+			options_dict["interleave"] = ["false","true"]
+			
+		self.metaattr_list = NVList(glade.get_widget("attributes"),
 					glade.get_widget("addattr"), 
 					glade.get_widget("delattr"),
 					options_dict,
@@ -1234,9 +1404,6 @@ class AddNativeDlg :
 				       glade.get_widget("addparam"),
 				       glade.get_widget("delparam"),
 				       _("Add Parameter"), fields)
-		param_list.insert({"id":uuid(),"name":"target_role",
-				"value":"stopped","shortdesc": \
-		_("press \"Default\" or \"Start\" button in toolbar/menu to start the resource")})
 		self.param_list = param_list
 		if meta != None :
 			for param in meta.parameters :
@@ -1424,21 +1591,58 @@ class AddNativeDlg :
 					msgbox(_("the Group Name is exists"))
 					passed = False
 
+				if group_ != "" and glade.get_widget("is_clone").get_active():
+					msgbox(_("Clone can't belong to group"))
+					passed = False
+				if group_ != "" and glade.get_widget("is_master").get_active():
+					msgbox(_("Master/Slave can't belong to group"))
+					passed = False
+
 				if passed :
 					rsc = {}
 					rsc["id"] = glade.get_widget("id").get_text()
 					(rsc["class"],rsc["type"],rsc["provider"]) = self.get_cur_select()
 					rsc["group"] = glade.get_widget("group").child.get_text()
+					if rsc["group"] != "" and not manager.rsc_exists(rsc["group"]):
+						rsc["new_group"] = "True"
+					else:
+						rsc["new_group"] = ""
 					rsc["params"] = self.param_list.get_data()
+					target_role_set = {}
+					param_names = []
 					for param in rsc["params"] :
 						if param["name"] == "target_role" :
 							param["id"] = rsc["id"]+"_target_role"
+							target_role_set = param
+						param_names.append(param["name"])
+					repeated_param = False
+					for param_name in param_names:
+						if param_names.count(param_name) > 1:
+							repeated_param = True
+							break
+					if repeated_param:
+						msgbox(_("Parameter \""+param_name+"\" is defined repeatedly"))
+						continue
 					if glade.get_widget("is_clone").get_active() :
 						rsc["advance"] = "clone"
 					elif glade.get_widget("is_master").get_active() :
 						rsc["advance"] = "master"
 					else :
 						rsc["advance"] = ""
+
+					if (rsc["group"] != "" or rsc["advance"] != "") and target_role_set != {}:
+						retval = confirmbox(_("Recommended:")+"\n"
+							+_("Select \"Clear\" to remove the set.")+"\n"
+							+_("It will inherit the attribute from its parent resource."),
+							(gtk.STOCK_CLEAR, gtk.RESPONSE_NO,
+							gtk.STOCK_YES, gtk.RESPONSE_YES,
+							gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL),
+							_("Are you sure you want to set \"target_role\"?"))
+						if retval == gtk.RESPONSE_CANCEL:
+							continue
+						if retval == gtk.RESPONSE_NO:
+							rsc["params"].remove(target_role_set)
+
 					rsc["advance_id"] = glade.get_widget("advance_id").get_text()
 					rsc["clone_max"] = glade.get_widget("clone_max").get_text()
 					rsc["clone_node_max"] = glade.get_widget("clone_node_max").get_text()
@@ -1485,7 +1689,7 @@ class MainWindow :
 		self.set_action_sensitive('delrsc',
 			manager.connected
 			and self.cur_type in [_("native"),_("group"),_("clone"),
-				_("place"),_("order"),_("colocation"),_("master")])
+				_("location"),_("order"),_("colocation"),_("master")])
 
 		self.set_action_sensitive('standby',
 			manager.connected
@@ -1532,7 +1736,7 @@ class MainWindow :
 	def can_change_view(self) :
 		if not self.cur_view.changed :
 			return True
-		ret = confirmbox("The data of current view has been changed. \nApply the change?",
+		ret = confirmbox(_("The data of current view have been changed.")+"\n"+_("Apply the changes?"),
 			(gtk.STOCK_YES, gtk.RESPONSE_YES, 
 			gtk.STOCK_NO, gtk.RESPONSE_NO,
 			gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
@@ -1555,8 +1759,8 @@ class MainWindow :
 			self.cur_view = NodeView(name)
 		elif type in [_("native")] :
 			self.cur_view = RscView(name)
-		elif type == _("place") :
-			self.cur_view = PlaceView(name)
+		elif type == _("location") :
+			self.cur_view = LocationView(name)
 		elif type == _("order") :
 			self.cur_view = OrderView(name)
 		elif type == _("colocation") :
@@ -1591,13 +1795,14 @@ class MainWindow :
 		server = glade.get_widget("server").get_text()
 		user = glade.get_widget("username").get_text()
 		password = glade.get_widget("password").get_text()
-		manager.save_login_info(server,user)
 		
 		dialog.destroy()
 
 		if ret == gtk.RESPONSE_OK:
 			if not manager.login(server, user, password):
 				msgbox(manager.failed_reason)
+			else:
+				manager.save_login_info(server,user)
 	
 		self.update_ui()
 
@@ -1617,9 +1822,17 @@ class MainWindow :
 			manager.do_cmd("standby\n"+self.cur_name + "\n" + "off")
 
 	def on_add_item(self, action) :
+		if self.cur_type in [_("constraints"), _("locations"), _("location")]:
+			add_type = _("location") 	
+		elif self.cur_type in [_("orders"), _("order")]:
+			add_type = _("order")
+		elif self.cur_type in [_("colocations"), _("colocation")]:
+			add_type = _("colocation")
+		else:
+			add_type = _("native")
 		new_type = kvbox(_("The type of new item"),
-				[Field("type",_("Item Type"),_("native"),
-				[_("native"), _("group"), _("place"),_("order"), _("colocation")])])
+				[Field("type",_("Item Type"),add_type,
+				[_("native"), _("group"), _("location"),_("order"), _("colocation")])])
 		if new_type == None :
 			return
 		if new_type["type"] == _("native"):
@@ -1638,14 +1851,15 @@ class MainWindow :
 				dlg = AddNativeDlg()
 				dlg.run()
 				
-		elif new_type["type"] == _("place") :
-			place = kvbox(_("Add Place Constraint"),
-				[Field("id",_("ID"),"place_",None, False),
-				 Field("rsc",_("Resource"),None, manager.get_all_rsc_id(),False),
-				 Field("score",_("Score"),"100")])
-			if place != None :
-				place["exprs"] = []
-				manager.update_constraint("rsc_location", place)
+		elif new_type["type"] == _("location") :
+			location = kvbox(_("Add Location Constraint"),
+				[Field("id",_("ID"),"location_",None, False),
+				 Field("rsc",_("Resource"),None, manager.get_all_rsc_id(),False)])
+			if location != None :
+				location["score"] = '0'
+				location["boolean_op"] = ''
+				location["exprs"] = []
+				manager.update_constraint("rsc_location", location)
 		elif new_type["type"] == _("order") :
 			order = kvbox(_("Add Order Constraint"),
 				[Field("id",_("ID"),"order_",None,False),
@@ -1678,7 +1892,7 @@ class MainWindow :
 					manager.do_cmd("del_rsc\n"+parent)
 				else :
 					manager.do_cmd("del_rsc\n"+self.cur_name)
-			elif self.cur_type == _("place") :
+			elif self.cur_type == _("location") :
 				manager.do_cmd("del_co\nrsc_location\n"+self.cur_name)
 			elif self.cur_type == _("order") :
 				manager.do_cmd("del_co\nrsc_order\n"+self.cur_name)
@@ -1695,15 +1909,36 @@ class MainWindow :
 		else :
 			target_role = "#default"
 				
-		(attrs, running_on, params, ops) = \
+		(attrs, running_on, metaattrs, params, ops) = \
 			manager.get_rsc_info(self.cur_name)
 		for param in params :
 			if param["name"] == "target_role" and param["value"] == target_role :
 				return
+
+		for subrsc in manager.get_rsc_sub_rsc(self.cur_name) :
+			(attrs, running_on, metaattrs, params, ops) = \
+				manager.get_rsc_info(subrsc)
+			for metaattr in metaattrs :
+				if metaattr["name"] == "target_role":
+					retval = confirmbox(_("Caused by your previous action, this attribute has been set to \"")
+								+metaattr["value"]
+								+_("\".")+"\n\n"
+								+_("Recommended:")+"\n"
+								+_("Select \"Clear\" to remove the set.")+"\n"
+								+_("The sub-resource will inherit this attribute from current resource."),
+								(gtk.STOCK_CLEAR, gtk.RESPONSE_NO,
+								gtk.STOCK_YES, gtk.RESPONSE_YES,
+								gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL),
+								_("Do you want to retain the \"target_role\" set for sub-resource \"")
+								+subrsc+_("\"?"))
+					if retval == gtk.RESPONSE_CANCEL:
+						return	
+					if retval == gtk.RESPONSE_NO:
+						new_metaattrs = metaattrs[:]
+						new_metaattrs.remove(metaattr)
+						manager.update_attrs("up_rsc_metaattrs\n"+subrsc ,"del_rsc_metaattr",
+							metaattrs, new_metaattrs, ["id","name","value"]);
 		manager.do_cmd("set_target_role\n%s\n%s"%(self.cur_name,target_role))
-		if target_role in ["started", "#default"] :
-			for subrsc in manager.get_rsc_sub_rsc(self.cur_name) :
-				manager.do_cmd("set_target_role\n%s\n%s"%(subrsc,target_role))
 	
 	def on_item_move(self, action) :
 		if action.get_name() == "uprsc" :
@@ -1926,11 +2161,16 @@ class Manager :
 		save_path = os.environ["HOME"]+"/.haclient"
 		if not os.path.exists(save_path) :
 			return ("127.0.0.1","hacluster")
-		return pickle.load(file(save_path,"r"))
-		
+		try:
+			return pickle.load(file(save_path,"r"))
+		except:
+			return ("127.0.0.1", "hacluster")	
 	def save_login_info(self, server, user) :
 		save_path = os.environ["HOME"]+"/.haclient"
-		pickle.dump((server,user), file(save_path,"w"))
+		try:
+			pickle.dump((server,user), file(save_path,"w"))
+		except:
+			os.remove(save_path)
 		return
 		
 	def login(self, server, username, password) :
@@ -2054,6 +2294,60 @@ class Manager :
 		self.update_timer = -1
 
 	# cluster functions		
+	def get_crm_metadata(self, crm_cmd) :
+		if crm_cmd == None :
+			return None
+		lines = self.query("crm_metadata\n%s"%(crm_cmd),True)
+
+		if lines == None :
+			return None
+		meta_data = ""
+		for line in lines :
+			if len(line)!= 0 :
+				meta_data = meta_data + line + "\n"
+
+		try :
+			doc_xml = parseString(meta_data).documentElement
+		except xml.parsers.expat.ExpatError:
+			debug("fail to parse the metadata of "+rsc_type)
+			return None
+			
+		meta = RAMeta()
+		meta.name = doc_xml.getAttribute("name")
+
+		meta.version = ""
+		version_xml = doc_xml.getElementsByTagName("version")
+		if version_xml != [] and version_xml[0] in doc_xml.childNodes :
+			meta.version = version_xml[0].childNodes[0].data
+
+		meta.longdesc = self.get_locale_desc(doc_xml, "longdesc");
+		meta.shortdesc = self.get_locale_desc(doc_xml, "shortdesc");
+		
+		meta.parameters = []
+		for param_xml in doc_xml.getElementsByTagName("parameter") :
+			param = {}
+			param["name"] = param_xml.getAttribute("name")
+			param["unique"] = param_xml.getAttribute("unique")
+			param["longdesc"] = self.get_locale_desc(param_xml, "longdesc");
+			param["shortdesc"] = self.get_locale_desc(param_xml, "shortdesc");
+			content_xml = param_xml.getElementsByTagName("content")[0]
+			content = {}
+			content["type"] = content_xml.getAttribute("type")
+			content["default"] = content_xml.getAttribute("default")
+			if content["type"] == "enum":
+				values_tag = "Allowed values:"
+				index = param["longdesc"].rfind(values_tag)
+				if index != -1:
+					strings = param["longdesc"][index+len(values_tag):].split(",")
+					content["values"] = []
+					for string in strings:
+						content["values"].append(string.strip())
+				
+			param["content"] = content
+			meta.parameters.append(param)
+		return meta
+
+
 	def get_cluster_config(self) :
 		
 		hb_attr_names = ["apiauth","auto_failback","baud","debug","debugfile",
@@ -2061,27 +2355,42 @@ class Manager :
 				 "initdead","keepalive","logfacility","logfile",
 				 "msgfmt","nice_failback","node","normalpoll",
 			 	"stonith","udpport","warntime","watchdog", "cluster"]
-		crm_attr_names = ["default-action-timeout","symmetric-cluster","stonith-enabled",
-				  "no-quorum-policy","default-resource-stickiness", "have_quorum",
-				  "default-resource-failure-stickiness"]
 				  
 		values = manager.query("hb_config")
 		if values == None :
 			return None
 		config = dict(zip(hb_attr_names,values))
 				
-		values = manager.query("crm_config")
-		if values == None :
-			return None
-		config.update(dict(zip(crm_attr_names,values)))
-		
+		value = manager.query("crm_config\nhave_quorum")
+		if value != None and value != []:
+			config["have_quorum"] = value[0]
+
+		return config
+
+	def get_crm_config(self, metadata):
+		config = {}
+		for parameter in metadata.parameters:
+			value = manager.query("crm_config\n%s"% \
+					str(parameter["name"]))
+			if value == None or value == []:
+				continue
+			if value == [""]:
+				config[parameter["name"]] = str(parameter["content"]["default"])
+			else:
+				config[parameter["name"]] = value[0]
 		return config
 		
-	def update_crm_config(self, new_crm_config) :
-		cur_crm_config = self.get_cluster_config()
+	def update_crm_config(self, metadata, new_crm_config) :
 		for k,v in new_crm_config.iteritems() :
-			if cur_crm_config[k] != v :
-				self.do_cmd("up_crm_config\n"+k+"\n"+v)
+			value = manager.query("crm_config\n"+str(k))
+			if value == [""]:
+				for parameter in metadata.parameters:
+					if parameter["name"] == k:
+						cur_value = str(parameter["content"]["default"])
+			else:
+				cur_value = value[0]
+			if cur_value != v :
+				self.do_cmd("up_crm_config\n"+str(k)+"\n"+v)
 				if self.failed_reason != "" :
 					msgbox(self.failed_reason)
 
@@ -2153,6 +2462,7 @@ class Manager :
 		cmd += "\n"+rsc["clone_node_max"]
 		cmd += "\n"+rsc["master_max"]
 		cmd += "\n"+rsc["master_node_max"]
+		cmd += "\n"+rsc["new_group"]
 		for param in rsc["params"] :
 			cmd += "\n"+param["id"]
 			cmd += "\n"+param["name"]
@@ -2164,11 +2474,17 @@ class Manager :
 		if group_id != None and group_id != "" :
 			ordered = self.cache_lookup(group_id+"\nordered")
 			if ordered != None :
-				self.update_rsc_attr(group_id, "ordered", ordered)
+				new_metaattrs = [{"id": group_id + "_metaattr_ordered" ,
+                                                "name": "ordered", "value": ordered}]
+				manager.update_attrs("up_rsc_metaattrs\n"+group_id,"del_rsc_metaattr",
+						[], new_metaattrs, ["id","name","value"]);
 				self.cache_delkey(group_id+"\nordered")
 			collocated = self.cache_lookup(group_id+"\ncollocated")
 			if collocated != None :
-				self.update_rsc_attr(group_id, "collocated", collocated)
+				new_metaattrs = [{"id": group_id + "_metaattr_collocated" ,
+                                                "name": "collocated", "value": collocated}]
+				manager.update_attrs("up_rsc_metaattrs\n"+group_id,"del_rsc_metaattr",
+						[], new_metaattrs, ["id","name","value"]);
 				self.cache_delkey(group_id+"\ncollocated")
 				
 	def get_all_rsc_id(self) :
@@ -2199,13 +2515,18 @@ class Manager :
 		op_attr_names = ["id", "name","description","interval","timeout",
 				 "start_delay","disabled","role","prereq","on_fail"]
 		param_attr_names = ["id", "name", "value"]
+		meta_attr_names = ["id", "name", "value"]
 		
 		attr_list = self.query("rsc_attrs\n%s"%rsc)
 		attrs = {}
 		if attr_list != None :
 			attrs = dict(zip(rsc_attr_names, attr_list))
+		
 		running_on = self.query("rsc_running_on\n%s"%rsc)
 
+		raw_metaattrs = self.query("rsc_metaattrs\n%s"%rsc)
+		metaattrs = self.split_attr_list(raw_metaattrs, meta_attr_names)
+
 		raw_params = self.query("rsc_params\n%s"%rsc)
 
 		params = self.split_attr_list(raw_params, param_attr_names)
@@ -2214,8 +2535,12 @@ class Manager :
 		raw_ops = raw_ops[1:]
 		ops = self.split_attr_list(raw_ops, op_attr_names)
 
-		return (attrs, running_on, params, ops)
+		return (attrs, running_on, metaattrs, params, ops)
 
+	def get_rsc_metaattrs(self, rsc) :
+		meta_attr_names = ["id", "name", "value"]
+		raw_metaattrs = self.query("rsc_metaattrs\n%s"%rsc)
+		return self.split_attr_list(raw_metaattrs, meta_attr_names)
 	def get_rsc_params(self, rsc) :
 		param_attr_names = ["id", "name", "value"]
 		raw_params = self.query("rsc_params\n%s"%rsc)
@@ -2268,6 +2593,8 @@ class Manager :
 				continue
 			if child.tagName != tag :
 				continue
+			if len(child.childNodes) == 0:
+				break
 			langtag = string.lower(child.getAttribute("lang"))
 			if langtag == "" :
 				desc_en = child.childNodes[0].data
@@ -2423,17 +2750,17 @@ class Manager :
 			
 	def get_constraint(self, type, id) :
 		if type == "rsc_location" :
-			place_attr_names = ["id","rsc","score"]
+			location_attr_names = ["id","rsc","score", "boolean_op"]
 			expr_attr_names = ["id","attribute","operation","value"]
 			attrs = manager.query("get_co\nrsc_location\n"+id)
 			if attrs == None :
 				return None
-			place = dict(zip(place_attr_names,attrs[:3]))
-			place["exprs"] = []
-			for i in range((len(attrs)-len(place_attr_names))/len(expr_attr_names)) :
-				expr = dict(zip(expr_attr_names,attrs[3+i*4:7+i*4]))
-				place["exprs"].append(expr)
-			return place
+			location = dict(zip(location_attr_names,attrs[:4]))
+			location["exprs"] = []
+			for i in range((len(attrs)-len(location_attr_names))/len(expr_attr_names)) :
+				expr = dict(zip(expr_attr_names,attrs[4+i*4:8+i*4]))
+				location["exprs"].append(expr)
+			return location
 		elif type == "rsc_order" :
 			order_attr_names = ["id","from","type","to"]
 			attrs = manager.query("get_co\nrsc_order\n" + id)
@@ -2457,8 +2784,8 @@ class Manager :
 			manager.do_cmd("del_co\nrsc_location\n"+constraint["id"])
 			if self.failed_reason != "" :
 				msgbox(self.failed_reason)
-			cmd = "up_co\nrsc_location\n%s\n%s\n%s"% \
-					(constraint["id"],constraint["rsc"],constraint["score"])
+			cmd = "up_co\nrsc_location\n%s\n%s\n%s\n%s"% \
+					(constraint["id"],constraint["rsc"],constraint["score"],constraint["boolean_op"])
 			for expr in constraint["exprs"] :
 				cmd = cmd + "\n" + expr["id"] + "\n" + expr["attribute"] + \
 					"\n" + expr["operation"] + "\n" + expr["value"]
